
/*
 * The Real SoundTracker - instrument editor
 *
 * Copyright (C) 1998-2019 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <glib/gi18n.h>
#include <stdlib.h>
#include <math.h>

#include "clavier.h"
#include "envelope-box.h"
#include "errors.h"
#include "extspinbutton.h"
#include "file-operations.h"
#include "gui-settings.h"
#include "gui-subs.h"
#include "gui.h"
#include "instrument-editor.h"
#include "keys.h"
#include "module-info.h"
#include "sample-editor.h"
#include "st-subs.h"
#include "track-editor.h"
#include "tracker.h"
#include "xm.h"

file_op *saveinstr, *loadinstr;

struct InstrumentEditor {
    gboolean full;
    guint idle_handler;
};

static GtkWidget *volenv, *panenv, *disableboxes[5];
static GtkWidget* instrument_editor_vibtype_w;
static GtkWidget* clavier;
static GtkWidget* curnote_label;
static GtkAdjustment* instrument_page_adj[4];

static STInstrument *current_instrument, *tmp_instrument = NULL;
static gint current_instrument_number = 0;

static void
instrument_page_volfade_changed(GtkAdjustment *adj)
{
    current_instrument->volfade = rint(gtk_adjustment_get_value(adj));
    gui_xm_set_modified(1);
}

static void
instrument_page_vibspeed_changed(GtkAdjustment *adj)
{
    current_instrument->vibrate = rint(gtk_adjustment_get_value(adj));
    gui_xm_set_modified(1);
}

static void
instrument_page_vibdepth_changed(GtkAdjustment *adj)
{
    current_instrument->vibdepth = rint(gtk_adjustment_get_value(adj));
    gui_xm_set_modified(1);
}

static void
instrument_page_vibsweep_changed(GtkAdjustment *adj)
{
    current_instrument->vibsweep = rint(gtk_adjustment_get_value(adj));
    gui_xm_set_modified(1);
}

static void
instrument_editor_vibtype_changed(GtkComboBox* combo)
{
    gint type = gtk_combo_box_get_active(combo);

    if (current_instrument->vibtype != type) {
        current_instrument->vibtype = type;
        gui_xm_set_modified(1);
    }
}

static gint
instrument_editor_clavierkey_press_event(GtkWidget* widget,
    gint key,
    gint button,
    gpointer data)
{
    if (button == 3) {
        current_instrument->samplemap[key] = gui_get_current_sample();
        gtk_widget_queue_draw(clavier);
        gui_xm_set_modified(1);
    } else {
        track_editor_do_the_note_key(key - 12 * gui_get_current_octave_value(), TRUE, 0, 0, TRUE);
        clavier_press(CLAVIER(clavier), key);
    }
    return FALSE;
}

static void
instrument_editor_init_samplemap(void)
{
    int key;
    int sample = gui_get_current_sample();

    for (key = 0; key < ARRAY_SIZE(current_instrument->samplemap); key++) {
        current_instrument->samplemap[key] = sample;
    }

    clavier_set_key_labels(CLAVIER(clavier), current_instrument->samplemap);
}

static gint
instrument_editor_clavierkey_release_event(GtkWidget* widget,
    gint key,
    gint button,
    gpointer data)
{
    if (button != 3) {
        track_editor_do_the_note_key(key - 12 * gui_get_current_octave_value(), FALSE, 0, 0, TRUE);
        clavier_release(CLAVIER(clavier), key);
    }
    return FALSE;
}

static gint
instrument_editor_clavierkey_enter_event(GtkWidget* widget,
    gint key,
    gpointer data)
{
    gtk_label_set_text(GTK_LABEL(curnote_label), instrument_editor_get_note_name(key));
    return FALSE;
}

static gint
instrument_editor_clavierkey_leave_event(GtkWidget* widget,
    gint key,
    gpointer data)
{
    gtk_label_set_text(GTK_LABEL(curnote_label), NULL);
    return FALSE;
}

static void
instrument_editor_load_instrument(const gchar* fn, const gchar* localname)
{
    STInstrument* instr = current_instrument;
    FILE* f;
    gboolean modified = FALSE;

    g_assert(instr != NULL);

    // Instead of locking the instrument and samples, we simply stop playing.
    gui_play_stop();

    f = fopen(localname, "rb");
    if (f) {
        statusbar_update(STATUS_LOADING_INSTRUMENT, TRUE);
        if (xm_load_xi(instr, f)) {
            statusbar_update(STATUS_INSTRUMENT_LOADED, FALSE);
            modified = TRUE;
        } else
            statusbar_update(STATUS_IDLE, FALSE);
        fclose(f);
    } else {
        static GtkWidget* dialog = NULL;
        gui_error_dialog(&dialog, _("Can't open file."), FALSE);
    }

    if (modified) {
        instrument_editor_update(TRUE);
        sample_editor_set_sample(&instr->samples[gui_get_current_sample()]);
        gui_xm_set_modified(TRUE);
    }
}

static void
instrument_editor_save_instrument(const gchar* fn, gchar* localname)
{
    STInstrument* instr = current_instrument;
    FILE* f;

    g_assert(instr != NULL);

    f = fopen(localname, "wb");

    if (f) {
        statusbar_update(STATUS_SAVING_INSTRUMENT, TRUE);
        if (xm_save_xi(instr, f)) {
            static GtkWidget* dialog = NULL;

            gui_error_dialog(&dialog, _("Saving instrument failed."), FALSE);
            statusbar_update(STATUS_IDLE, FALSE);
        } else
            statusbar_update(STATUS_INSTRUMENT_SAVED, FALSE);
        fclose(f);
    } else {
        static GtkWidget* dialog = NULL;
        gui_error_dialog(&dialog, _("Can't open file."), FALSE);
    }
}

const gchar*
instrument_editor_get_note_name(const guint key)
{
    guint index = (gui_settings.sharp ? 0 : 1) + (gui_settings.bh ? 2 : 0);
    g_assert(key < NUM_NOTES);
    return notenames[index][key];
}

gboolean instrument_editor_clear_current_instrument(void)
{
    if (GTK_TOGGLE_BUTTON(editing_toggle)->active) {
        if (gui_ok_cancel_modal(mainwindow,
            _("Clear current instrument?"))) {
            gui_play_stop();

            st_clean_instrument(current_instrument, NULL);

            instrument_editor_update(TRUE);
            sample_editor_update();
            gui_xm_set_modified(TRUE);
        }
        return TRUE;
    }
    return FALSE;
}

void instrument_page_create(GtkNotebook* nb)
{
    GtkWidget *mainbox, *vbox, *thing, *box, *box2, *box3, *box4, *frame, *table;
    GtkListStore* list_store;
    GtkTreeIter iter;
    guint i;
    static const char* vibtypelabels[] = { N_("Sine"), N_("Square"), N_("Saw Down"), N_("Saw Up"), NULL };

    static const gchar* xi_f[] = { N_("FastTracker instruments (*.xi)"), "*.[xX][iI]", NULL };
    static const gchar** formats[] = { xi_f, NULL };

    mainbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(mainbox), 10);
    gtk_notebook_append_page(nb, mainbox, gtk_label_new(_("Instrument Editor")));
    gtk_widget_show(mainbox);

    add_empty_vbox(mainbox);

    disableboxes[0] = vbox = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, TRUE, 0);
    gtk_widget_show(vbox);

    volenv = envelope_box_new(_("Volume envelope"));
    gtk_box_pack_start(GTK_BOX(vbox), volenv, TRUE, TRUE, 0);
    gtk_widget_show(volenv);

    thing = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, TRUE, 0);
    gtk_widget_show(thing);

    panenv = envelope_box_new(_("Panning envelope"));
    gtk_box_pack_start(GTK_BOX(vbox), panenv, TRUE, TRUE, 0);
    gtk_widget_show(panenv);

    thing = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(mainbox), thing, FALSE, TRUE, 0);
    gtk_widget_show(thing);

    box = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(mainbox), box, FALSE, TRUE, 0);
    gtk_widget_show(box);

    box2 = gtk_vbox_new(TRUE, 2);
    gtk_box_pack_start(GTK_BOX(box), box2, TRUE, TRUE, 0);
    gtk_widget_show(box2);

    loadinstr = fileops_dialog_create(DIALOG_LOAD_INSTRUMENT, _("Load Instrument"),
        &gui_settings.loadinstr_path, instrument_editor_load_instrument, TRUE, FALSE, formats,
        N_("Load instrument in the current instrument slot"), NULL);
    saveinstr = fileops_dialog_create(DIALOG_SAVE_INSTRUMENT, _("Save Instrument"),
        &gui_settings.saveinstr_path, instrument_editor_save_instrument, TRUE, TRUE, formats,
        N_("Save the current instrument"), "xi");

    thing = gtk_button_new_with_label(_("Load XI"));
    gtk_box_pack_start(GTK_BOX(box2), thing, TRUE, TRUE, 0);
    gtk_widget_show(thing);
    g_signal_connect_swapped(thing, "clicked",
        G_CALLBACK(fileops_open_dialog), loadinstr);

    disableboxes[3] = thing = gtk_button_new_with_label(_("Save XI"));
    gtk_box_pack_start(GTK_BOX(box2), thing, TRUE, TRUE, 0);
    gtk_widget_show(thing);
    g_signal_connect_swapped(thing, "clicked",
        G_CALLBACK(fileops_open_dialog), saveinstr);

    thing = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(box), thing, FALSE, TRUE, 0);
    gtk_widget_show(thing);

    disableboxes[1] = box2 = gtk_vbox_new(TRUE, 2);
    gtk_box_pack_start(GTK_BOX(box), box2, TRUE, TRUE, 0);

    box3 = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(box2), box3, FALSE, TRUE, 0);
    thing = gtk_label_new(_("Volume Fadeout"));
    gtk_box_pack_start(GTK_BOX(box3),thing, FALSE, FALSE, 0);
    instrument_page_adj[0] = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0xfff, 1.0, 400.0, 0.0));
    thing = extspinbutton_new(instrument_page_adj[0], 0.0, 0, TRUE);
    gtk_box_pack_end(GTK_BOX(box3),thing, FALSE, FALSE, 0);
    thing = gtk_hscale_new(instrument_page_adj[0]);
    gtk_scale_set_draw_value(GTK_SCALE(thing), FALSE);
    gtk_box_pack_start(GTK_BOX(box2), thing, FALSE, TRUE, 0);
    gtk_widget_show_all(box2);
    g_signal_connect(instrument_page_adj[0], "value_changed",
        G_CALLBACK(instrument_page_volfade_changed), NULL);

    thing = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(box), thing, FALSE, TRUE, 0);
    gtk_widget_show(thing);

    disableboxes[4] = table = gtk_table_new(2, 2, TRUE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 8);
    gtk_box_pack_start(GTK_BOX(box), table, TRUE, TRUE, 0);

    box3 = gtk_hbox_new(FALSE, 4);
    gtk_table_attach_defaults(GTK_TABLE(table), box3, 0, 1, 0, 1);

    thing = gtk_label_new(_("Vibrato:"));
    gtk_box_pack_start(GTK_BOX(box3), thing, FALSE, FALSE, 0);
    list_store = gtk_list_store_new(1, G_TYPE_STRING);
    for (i = 0; vibtypelabels[i] != NULL; i++) {
        gtk_list_store_append(list_store, &iter);
        gtk_list_store_set(list_store, &iter, 0, _(vibtypelabels[i]), -1);
    }
    instrument_editor_vibtype_w = thing = gui_combo_new(list_store);
    g_signal_connect(thing, "changed", G_CALLBACK(instrument_editor_vibtype_changed), NULL);
    gtk_box_pack_end(GTK_BOX(box3), thing, FALSE, TRUE, 0);
    thing = gtk_label_new(_("Type"));
    gtk_box_pack_end(GTK_BOX(box3), thing, FALSE, FALSE, 0);
    thing = gui_subs_create_slider(_("Speed"), 0, 0x3f, instrument_page_vibspeed_changed, &instrument_page_adj[1], TRUE);
    gtk_table_attach_defaults(GTK_TABLE(table), thing, 1, 2, 0, 1);

    thing = gui_subs_create_slider(_("Depth"), 0, 0xf, instrument_page_vibdepth_changed, &instrument_page_adj[2], TRUE);
    gtk_table_attach_defaults(GTK_TABLE(table), thing, 0, 1, 1, 2);
    thing = gui_subs_create_slider(_("Sweep"), 0, 0xff, instrument_page_vibsweep_changed, &instrument_page_adj[3], TRUE);
    gtk_table_attach_defaults(GTK_TABLE(table), thing, 1, 2, 1, 2);
    gtk_widget_show_all(table);

    thing = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(mainbox), thing, FALSE, TRUE, 0);
    gtk_widget_show(thing);

    // Sample map editor coming up
    disableboxes[2] = box2 = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(mainbox), box2, FALSE, TRUE, 0);
    gtk_widget_show(box2);

    box3 = gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(box2), box3, TRUE, TRUE, 0);
    gtk_widget_show(box3);

    box = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(box), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_show(box);
    gtk_box_pack_start(GTK_BOX(box3), box, TRUE, TRUE, 0);

    clavier = clavier_new();
    gtk_widget_show(clavier);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(box), clavier);

    gtk_widget_set_size_request(clavier, NUM_NOTES * 7, 50);

    clavier_set_clavier_type(CLAVIER(clavier), CLAVIER_TYPE_SEQUENCER);
    clavier_set_range(CLAVIER(clavier), 0, 95);
    clavier_set_show_middle_c(CLAVIER(clavier), FALSE);
    clavier_set_show_transpose(CLAVIER(clavier), FALSE);

    g_signal_connect(clavier, "clavierkey_press",
        G_CALLBACK(instrument_editor_clavierkey_press_event),
        NULL);
    g_signal_connect(clavier, "clavierkey_release",
        G_CALLBACK(instrument_editor_clavierkey_release_event),
        NULL);
    g_signal_connect(clavier, "clavierkey_enter",
        G_CALLBACK(instrument_editor_clavierkey_enter_event),
        NULL);
    g_signal_connect(clavier, "clavierkey_leave",
        G_CALLBACK(instrument_editor_clavierkey_leave_event),
        NULL);

    box3 = gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(box2), box3, FALSE, TRUE, 0);
    gtk_widget_show(box3);

    add_empty_vbox(box3);

    thing = gtk_label_new(_("Note:"));
    gtk_widget_show(thing);
    gtk_box_pack_start(GTK_BOX(box3), thing, FALSE, TRUE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(box3), frame, FALSE, TRUE, 0);
    gtk_widget_show(frame);

    box4 = gtk_vbox_new(FALSE, 2);
    gtk_widget_show(box4);
    gtk_container_add(GTK_CONTAINER(frame), box4);
    gtk_container_set_border_width(GTK_CONTAINER(box4), 4);

    curnote_label = thing = gtk_label_new("");
    gtk_widget_show(thing);
    gtk_box_pack_start(GTK_BOX(box4), thing, FALSE, TRUE, 0);

    thing = gtk_button_new_with_label(_("Initialize"));
    gtk_widget_show(thing);
    gtk_box_pack_start(GTK_BOX(box3), thing, FALSE, TRUE, 0);
    g_signal_connect(thing, "clicked",
        G_CALLBACK(instrument_editor_init_samplemap), NULL);

    add_empty_vbox(box3);

    add_empty_vbox(mainbox);
}

void instrument_editor_copy_instrument(STInstrument* instr)
{
    if (tmp_instrument == NULL)
        tmp_instrument = calloc(1, sizeof(STInstrument));
    st_copy_instrument(instr, tmp_instrument);
}

void instrument_editor_cut_instrument(STInstrument* instr)
{
    instrument_editor_copy_instrument(instr);
    st_clean_instrument(instr, NULL);
}

void instrument_editor_paste_instrument(STInstrument* instr)
{
    if (tmp_instrument == NULL)
        return;
    st_copy_instrument(tmp_instrument, instr);
}

void instrument_editor_set_instrument(STInstrument* i, const gint ins)
{
    current_instrument = i;
    current_instrument_number = ins;

    instrument_editor_update(FALSE);
}

STInstrument*
instrument_editor_get_instrument(void)
{
    return current_instrument;
}

gboolean
instrument_editor_handle_keys(int shift,
    int ctrl,
    int alt,
    guint32 keyval,
    gboolean pressed)
{
    int i;

    i = keys_get_key_meaning(keyval, ENCODE_MODIFIERS(shift, ctrl, alt), -1);
    if (i != -1 && KEYS_MEANING_TYPE(i) == KEYS_MEANING_NOTE) {
        gboolean real_press = track_editor_do_the_note_key(i, pressed, keyval,
            ENCODE_MODIFIERS(shift, ctrl, alt), TRUE);
        if (real_press)
            (pressed ? clavier_press : clavier_release)
                (CLAVIER(clavier), i + 12 * gui_get_current_octave_value());
        return TRUE;
    }

    return FALSE;
}

static gboolean
instrument_editor_update_idle_function(struct InstrumentEditor *ie)
{
    gint n;
    gboolean o = current_instrument != NULL && st_instrument_num_samples(current_instrument) > 0;

    for (n = 0; n < ARRAY_SIZE(disableboxes); n++)
        gtk_widget_set_sensitive(disableboxes[n], o);

    if (!o) {
        envelope_box_set_envelope(ENVELOPE_BOX(volenv), NULL);
        envelope_box_set_envelope(ENVELOPE_BOX(panenv), NULL);
        clavier_set_key_labels(CLAVIER(clavier), NULL);
    } else {
        gboolean m = xm_get_modified();

        envelope_box_set_envelope(ENVELOPE_BOX(volenv), &current_instrument->vol_env);
        envelope_box_set_envelope(ENVELOPE_BOX(panenv), &current_instrument->pan_env);

        gtk_adjustment_set_value(instrument_page_adj[0], current_instrument->volfade);
        gtk_adjustment_set_value(instrument_page_adj[1], current_instrument->vibrate);
        gtk_adjustment_set_value(instrument_page_adj[2], current_instrument->vibdepth);
        gtk_adjustment_set_value(instrument_page_adj[3], current_instrument->vibsweep);
        gtk_combo_box_set_active(GTK_COMBO_BOX(instrument_editor_vibtype_w), current_instrument->vibtype);

        clavier_set_key_labels(CLAVIER(clavier), current_instrument->samplemap);

        gui_xm_set_modified(m);
    }

    if (ie->full)
        modinfo_update_instrument(current_instrument_number);

    ie->idle_handler = 0;
    return FALSE;
}

void
instrument_editor_update(gboolean full)
{
    static struct InstrumentEditor ie = {0};

    ie.full = full;

    if (current_instrument){
        gui_block_instrname_entry(TRUE); /* Preventing callback when changing instrument name entry */
        gtk_entry_set_text(GTK_ENTRY(gui_curins_name), current_instrument->utf_name);
        gui_block_instrname_entry(FALSE);
    }

    if (!ie.idle_handler) {
        ie.idle_handler = g_idle_add((GSourceFunc)instrument_editor_update_idle_function,
                                      &ie);
        g_assert(ie.idle_handler != 0);
    }
}
